@@ -22,6 +22,7 @@
  */
 
 #include "vncviewer.h"
+#include <ctype.h>
 #include <X11/Xaw/Toggle.h>
 
 char *programName;
@@ -30,6 +31,9 @@ Display* dpy;
 
 Widget toplevel;
 
+extern void raiseme(int force);
+extern void CreateChat(void);
+
 void set_sbwidth(int sbw) {
 	char *q, *p, t[5];
 	int i, k, N = 4;
@@ -66,6 +70,20 @@ void set_sbwidth(int sbw) {
 	}
 }
 
+void min_title(void) {
+	char *q;
+	int k;
+
+	k = 0;
+	while (fallback_resources[k] != NULL) {
+		q = strstr(fallback_resources[k], "Ssvnc.title: ");
+		if (q) {
+			fallback_resources[k] = strdup("Ssvnc.title: %s");
+		}
+		k++;
+	}
+}
+
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
@@ -75,7 +93,7 @@ void unixpw(char *instr, int vencrypt_plain) {
 	FILE *in;
 	int i, rmfile = 0;
 	struct stat sb;
-	int N = 99;
+	int N;
 	char username[100], passwd[100];
 	static int did = 0;
 
@@ -84,10 +102,16 @@ void unixpw(char *instr, int vencrypt_plain) {
 	}
 	did = 1;
 
-	for (i=0; i<100; i++) {
+	for (i=0; i < sizeof(username); i++) {
 		username[i] = '\0';
+	}
+	for (i=0; i < sizeof(passwd); i++) {
 		passwd[i] = '\0';
 	}
+	N = sizeof(username)-1; /* 99 */
+	if (sizeof(passwd)-1 < N) {
+		N = sizeof(passwd)-1;
+	}
 
 	if (instr == NULL) {
 		return;
@@ -105,11 +129,23 @@ void unixpw(char *instr, int vencrypt_plain) {
 	}
 	if (!strcmp(str, ".")) {
 		char *p;
-		fprintf(stderr, "\nUnix Username: ");
-		if (fgets(username, N, stdin) == NULL) {
-			exit(1);
+		if (!use_tty()) {
+			char *u;
+			fprintf(stderr, "\nEnter Unix Username and Password in the popups.\n");
+			u = DoUserDialog();
+			if (strlen(u) >= N) {
+				exit(1);
+			}
+			strncpy(username, u, N);
+			p = DoPasswordDialog();
+		} else {
+			raiseme(1);
+			fprintf(stderr, "\nUnix Username: ");
+			if (fgets(username, N, stdin) == NULL) {
+				exit(1);
+			}
+			p = getpass("Unix Password: ");
 		}
-		p = getpass("Unix Password: ");
 		if (! p) {
 			exit(1);
 		}
@@ -118,7 +154,13 @@ void unixpw(char *instr, int vencrypt_plain) {
 		
 	} else if (!strcmp(str, "-")) {
 		char *p, *q;
-		p = getpass("unixuser@unixpasswd: ");
+		if (!use_tty()) {
+			fprintf(stderr, "\nEnter unixuser@unixpasswd in the popup.\n");
+			p = DoPasswordDialog();
+		} else {
+			raiseme(1);
+			p = getpass("unixuser@unixpasswd: ");
+		}
 		if (! p) {
 			exit(1);
 		}
@@ -228,15 +270,21 @@ static void chat_window_only(void) {
 	}
 }
 
+int saw_appshare = 0;
+
 int
 main(int argc, char **argv)
 {
-	int i, save_sbw;
+	int i, save_sbw, saw_listen = 0;
 	char *pw_loc = NULL;
 	programName = argv[0];
 
+	if (strrchr(programName, '/') != NULL) {
+		programName = strrchr(programName, '/') + 1;
+	}
+
 	for (i = 1; i < argc; i++) {
-		if (strcmp(argv[i], "-env") == 0) {
+		if (!strcmp(argv[i], "-env")) {
 			if (i+1 < argc) {
 				char *estr = argv[i+1];
 				if (strchr(estr, '=')) {
@@ -244,6 +292,18 @@ main(int argc, char **argv)
 				}
 			}
 		}
+		if (!strcmp(argv[i], "-noipv4")) {
+			putenv("VNCVIEWER_NO_IPV4=1");
+		}
+		if (!strcmp(argv[i], "-noipv6")) {
+			putenv("VNCVIEWER_NO_IPV6=1");
+		}
+	}
+	if (getenv("VNCVIEWER_NO_IPV4")) {
+		appData.noipv4 = True;
+	}
+	if (getenv("VNCVIEWER_NO_IPV6")) {
+		appData.noipv6 = True;
 	}
 
   /* The -listen option is used to make us a daemon process which listens for
@@ -257,28 +317,49 @@ main(int argc, char **argv)
      flag. */
 
 	for (i = 1; i < argc; i++) {
+		if (!strcmp(argv[i], "-appshare")) {
+			putenv("SSVNC_MULTIPLE_LISTEN=1");
+			fprintf(stderr, "Enabling -multilisten mode for 'x11vnc -appshare' usage.\n\n");
+			saw_appshare = 1;
+		}
+		if (!strcmp(argv[i], "-multilisten")) {
+			putenv("SSVNC_MULTIPLE_LISTEN=1");
+			saw_listen = 2;
+		}
+		if (!strcmp(argv[i], "-listen")) {
+			saw_listen = 1;
+		}
+		if (!strcmp(argv[i], "-acceptpopup")) {
+			putenv("SSVNC_ACCEPT_POPUP=1");
+		}
+		if (!strcmp(argv[i], "-acceptpopupsc")) {
+			putenv("SSVNC_ACCEPT_POPUP_SC=1");
+		}
 		if (strstr(argv[i], " pw=") != NULL) {
 			pw_loc = strstr(argv[i], " pw=") + 1;
 		}
 	}
 
 	for (i = 1; i < argc; i++) {
-		if (strcmp(argv[i], "-listen") == 0) {
+		if (!strcmp(argv[i], "-appshare") && !saw_listen) {
 			listenForIncomingConnections(&argc, argv, i);
 			break;
 		}
-		if (strcmp(argv[i], "-multilisten") == 0) {
-			putenv("SSVNC_MULTIPLE_LISTEN=1");
+		if (!strcmp(argv[i], "-multilisten")) {
 			listenForIncomingConnections(&argc, argv, i);
 			break;
 		}
-		if (strcmp(argv[i], "-tunnel") == 0 || strcmp(argv[i], "-via") == 0) {
+		if (!strcmp(argv[i], "-listen")) {
+			listenForIncomingConnections(&argc, argv, i);
+			break;
+		}
+		if (!strcmp(argv[i], "-tunnel") || !strcmp(argv[i], "-via")) {
 			if (!createTunnel(&argc, argv, i)) {
 				exit(1);
 			}
 			break;
 		}
-		if (strcmp(argv[i], "-printres") == 0 || strcmp(argv[i], "-res") == 0) {
+		if (!strcmp(argv[i], "-printres") || !strcmp(argv[i], "-res")) {
 			int j = 0;
 			fprintf(stdout, "\n! Ssvnc fallback X resources:\n\n");
 			while (1) {
@@ -308,6 +389,9 @@ main(int argc, char **argv)
      generating appropriate resource specs, and makes a connection to the X
      display. */
 
+	if (saw_appshare || getenv("VNCVIEWER_MIN_TITLE")) {
+		min_title();
+	}
 	appData.sbWidth = 0;
 	if (getenv("VNCVIEWER_SBWIDTH")) {
 		int sbw = atoi(getenv("VNCVIEWER_SBWIDTH"));
@@ -346,6 +430,10 @@ main(int argc, char **argv)
 
 	GetArgsAndResources(argc, argv);
 
+	if (saw_appshare) {
+		appData.appShare = True;
+	}
+
 	if (save_sbw) {
 		appData.sbWidth = save_sbw;
 	}
@@ -391,7 +479,7 @@ main(int argc, char **argv)
 						tmp[i] = '\0';
 					}
 					for (i=0; i<250; i++) {
-						if (i >= strlen(appData.repeaterUltra)) {
+						if (i >= (int) strlen(appData.repeaterUltra)) {
 							break;
 						}
 						tmp[i] = appData.repeaterUltra[i];
@@ -422,6 +510,7 @@ main(int argc, char **argv)
 
 	CreatePopup();
 	CreateScaleN();
+	CreateTurboVNC();
 	CreateQuality();
 	CreateCompress();
 	CreateChat();
@@ -490,7 +579,7 @@ static Widget bfull = NULL;
 
 int do_format_change = 0;
 int do_cursor_change = 0;
-int do_fb_update = 0.0;
+double do_fb_update = 0.0;
 static void schedule_format_change(void) {
 	do_format_change = 1;
 	do_cursor_change = 0;
@@ -543,6 +632,7 @@ Toggle8bpp(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		fprintf(stderr, "8bpp: on (%d colors)\n", appDataNew.useBGR233);
 	}
 	schedule_format_change();
+	if (w || ev || params || num_params) {}
 }
 
 
@@ -563,6 +653,7 @@ Toggle16bpp(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		fprintf(stderr, "16bpp: on\n");
 	}
 	schedule_format_change();
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -590,6 +681,7 @@ ToggleFullColor(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		}
 	}
 	schedule_format_change();
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -603,6 +695,7 @@ ToggleXGrab(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	fprintf(stderr, "ToggleXGrab, current=%d\n", appData.grabAll);
 	/* always ungrab to be sure, fullscreen will handle the rest */
 	XUngrabServer(dpy);
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -613,6 +706,7 @@ ToggleEscapeActive(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		appData.escapeActive = True;
 	}
+	if (w || ev || params || num_params) {}
 }
 
 /*
@@ -635,6 +729,7 @@ Toggle256Colors(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		appDataNew.useBGR565 = False;
 		schedule_format_change();
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -649,6 +744,7 @@ Toggle64Colors(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		appDataNew.useBGR565 = False;
 		schedule_format_change();
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -663,6 +759,7 @@ Toggle8Colors(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		appDataNew.useBGR565 = False;
 		schedule_format_change();
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -678,6 +775,7 @@ ToggleGreyScale(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		fprintf(stderr, "greyscale: on\n");
 	}
 	schedule_format_change();
+	if (w || ev || params || num_params) {}
 }
 
 /*
@@ -696,6 +794,7 @@ ToggleJPEG(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		fprintf(stderr, "JPEG: on\n");
 	}
 	schedule_format_change();
+	if (w || ev || params || num_params) {}
 }
 
 /*
@@ -704,6 +803,7 @@ ToggleJPEG(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 
 static Bool usingZRLE = False;
 static Bool usingZYWRLE = False;
+static Bool usingHextile = False;
 extern int skip_maybe_sync;
 
 void
@@ -712,6 +812,7 @@ ToggleTightZRLE(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	char prefTight[] = "copyrect tight zrle zywrle zlib hextile corre rre raw";
 	char prefZRLE[]  = "copyrect zrle zywrle tight zlib hextile corre rre raw";
 	init_format_change();
+	usingHextile = False;
 	if (! appData.encodingsString) {
 		appDataNew.encodingsString = strdup(prefZRLE);
 		usingZRLE = True;
@@ -751,6 +852,7 @@ ToggleTightZRLE(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		first = 0;
 	}
 	schedule_format_change();
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -760,6 +862,7 @@ ToggleZRLEZYWRLE(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	char prefZRLE[]   = "copyrect zrle zywrle tight zlib hextile corre rre raw";
 	init_format_change();
 	usingZRLE = True;
+	usingHextile = False;
 	if (! appData.encodingsString) {
 		appDataNew.encodingsString = strdup(prefZYWRLE);
 		usingZYWRLE = True;
@@ -780,6 +883,54 @@ ToggleZRLEZYWRLE(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		}
 	}
 	schedule_format_change();
+	if (w || ev || params || num_params) {}
+}
+
+void
+ToggleTightHextile(Widget w, XEvent *ev, String *params, Cardinal *num_params)
+{
+	char prefTight[]   = "copyrect tight zrle zywrle zlib hextile corre rre raw";
+	char prefHextile[] = "copyrect hextile tight zrle zywrle zlib corre rre raw";
+	init_format_change();
+	usingZRLE = False;
+	usingZYWRLE = False;
+	if (! appData.encodingsString) {
+		appDataNew.encodingsString = strdup(prefHextile);
+		usingHextile = True;
+		fprintf(stderr, "prefer: Hextile\n");
+	} else {
+		char *t, *z;
+		static int first = 1;
+		t = strstr(appData.encodingsString, "tight");
+		z = strstr(appData.encodingsString, "hextile");
+		if (first && usingHextile) {
+			appDataNew.encodingsString = strdup(prefTight);
+			usingHextile = False;
+		} else if (! t) {
+			appDataNew.encodingsString = strdup(prefHextile);
+			usingHextile = True;
+			fprintf(stderr, "prefer: Hextile\n");
+		} else if (! z) {
+			appDataNew.encodingsString = strdup(prefTight);
+			usingHextile = False;
+			skip_maybe_sync = 0;
+			fprintf(stderr, "prefer: Tight\n");
+		} else {
+			if (t < z) {
+				appDataNew.encodingsString = strdup(prefHextile);
+				usingHextile = True;
+				fprintf(stderr, "prefer: Hextile\n");
+			} else {
+				appDataNew.encodingsString = strdup(prefTight);
+				usingHextile = False;
+				skip_maybe_sync = 0;
+				fprintf(stderr, "prefer: Tight\n");
+			}
+		}
+		first = 0;
+	}
+	schedule_format_change();
+	if (w || ev || params || num_params) {}
 }
 
 void scale_check_zrle(void) {
@@ -791,8 +942,8 @@ void scale_check_zrle(void) {
 	if (getenv("SSVNC_PRESERVE_ENCODING")) {
 		return;
 	}
-	if (!usingZRLE) {
-		Widget w;
+	if (!usingZRLE && !usingHextile) {
+		Widget w = 0;
 		fprintf(stderr, "\nSwitching to faster ZRLE encoding in client-side scaling mode.\n");
 		fprintf(stderr, "Switch back to Tight via the Popup menu if you prefer it.\n\n");
 		ToggleTightZRLE(w, NULL, NULL, NULL);
@@ -814,6 +965,7 @@ ToggleViewOnly(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		fprintf(stderr, "viewonly: on\n");
 	}
 	Xcursors(1);
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -833,6 +985,7 @@ ToggleCursorShape(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		do_cursor_change = -1;
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -845,6 +998,7 @@ ToggleCursorAlpha(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		appData.useCursorAlpha = True;
 		fprintf(stderr, "useCursorAlpha: on\n");
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -860,6 +1014,7 @@ ToggleX11Cursor(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	}
 	schedule_format_change();
 	do_cursor_change = 1;
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -872,6 +1027,7 @@ ToggleBell(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		appData.useBell = True;
 		fprintf(stderr, "useBell: on\n");
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -886,6 +1042,7 @@ ToggleRawLocal(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		fprintf(stderr, "useRawLocal: on\n");
 	}
 	schedule_format_change();
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -900,8 +1057,50 @@ ToggleServerInput(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		fprintf(stderr, "serverInput: on\n");
 		SendServerInput(False);
 	}
+	if (w || ev || params || num_params) {}
 }
 
+void
+TogglePipelineUpdates(Widget w, XEvent *ev, String *params, Cardinal *num_params)
+{
+	if (appData.pipelineUpdates) {
+		appData.pipelineUpdates= False;
+		fprintf(stderr, "pipeline-update: off\n");
+	} else {
+		appData.pipelineUpdates = True;
+		fprintf(stderr, "pipeline-update: on\n");
+	}
+	/* XXX request one to be sure? */
+	if (w || ev || params || num_params) {}
+}
+
+void
+ToggleSendClipboard(Widget w, XEvent *ev, String *params, Cardinal *num_params)
+{
+	if (appData.sendClipboard) {
+		appData.sendClipboard= False;
+		fprintf(stderr, "Send CLIPBOARD Selection: off (send PRIMARY instead)\n");
+	} else {
+		appData.sendClipboard = True;
+		fprintf(stderr, "Send CLIPBOARD Selection: on (do not send PRIMARY)\n");
+	}
+	if (w || ev || params || num_params) {}
+}
+
+void
+ToggleSendAlways(Widget w, XEvent *ev, String *params, Cardinal *num_params)
+{
+	if (appData.sendAlways) {
+		appData.sendAlways= False;
+		fprintf(stderr, "Send Selection Always: off\n");
+	} else {
+		appData.sendAlways = True;
+		fprintf(stderr, "Send Selection Always: on\n");
+	}
+	if (w || ev || params || num_params) {}
+}
+
+
 Bool _sw1_ = False;	/* XXX this is a weird bug... */
 Bool _sw2_ = False;
 Bool _sw3_ = False;
@@ -924,20 +1123,26 @@ ToggleSingleWindow(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 			XDefineCursor(dpy, desktopWin, bogoCursor);
 		}
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void raiseme(int force);
 void AppendChatInput(char *);
 
+extern void ShowChat(Widget w, XEvent *event, String *params, Cardinal *num_params);
+extern void ShowFile(Widget w, XEvent *event, String *params, Cardinal *num_params);
+extern Bool SendTextChatFinished(void);
+
+
 void printChat(char *str, Bool raise) {
 	if (appData.termChat) {
 		if (raise) {
 			raiseme(0);
 		}
-		fprintf(stderr, str);
+		fprintf(stderr, "%s", str);
 	} else {
 		if (raise) {
-			ShowChat();
+			ShowChat(0, 0, 0, 0);
 		}
 		AppendChatInput(str);
 	}
@@ -953,7 +1158,7 @@ ToggleTextChat(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		HideChat(0, NULL, NULL, NULL);
 		appData.chatActive= False;
 	} else {
-		ShowChat();
+		ShowChat(0, 0, 0, 0);
 		SendTextChatOpen();
 		if (appData.termChat) {
 			printChat("\n*SentOpen*\n\nSend: ", True);
@@ -962,6 +1167,7 @@ ToggleTextChat(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		}
 		appData.chatActive = True;
 	}
+	if (w || ev || params || num_params) {}
 }
 
 extern int filexfer_sock;
@@ -976,8 +1182,10 @@ ToggleFileXfer(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 {
 	static double last_start = 0.0;
 	if (appData.fileActive) {
-		//HideFile(w, ev, params, num_params);
-		//appData.fileActive = False;
+#if 0
+		HideFile(w, ev, params, num_params);
+		appData.fileActive = False;
+#endif
 #ifndef KILLJAVA
 		if (filexfer_sock >= 0) {
 			close(filexfer_sock);
@@ -1004,9 +1212,11 @@ ToggleFileXfer(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		appData.fileActive = True;
 		last_start = dnow();
 	}
+	if (w || ev || params || num_params) {}
 }
 
 static int fooHandler(Display *dpy, XErrorEvent *error) {
+	if (dpy || error) {}
 	return 0;
 }
 
@@ -1060,6 +1270,7 @@ DoServerScale(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		n = atoi(s);
 		set_server_scale(n);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void set_server_quality(int n) {
@@ -1094,6 +1305,7 @@ DoServerQuality(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		n = atoi(s);
 		set_server_quality(n);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void set_server_compress(int n) {
@@ -1128,9 +1340,11 @@ DoServerCompress(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		n = atoi(s);
 		set_server_compress(n);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 extern void rescale_image(void);
+extern void get_scale_values(double *fx, double *fy);
 
 void
 SetScale(Widget w, XEvent *ev, String *params, Cardinal *num_params)
@@ -1138,12 +1352,16 @@ SetScale(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	char *s;
 	s = DoScaleDialog();
 	if (s[0] != '\0') {
+#if 0
 		int w = si.framebufferWidth;
 		int h = si.framebufferHeight;
+#endif
+		double fx, fy;
 		int fs = 0;
 		if (appData.scale != NULL && !strcmp(s, appData.scale)) {
 			return;
 		}
+		
 		if (!strcasecmp(s, "none")) {
 			appData.scale = NULL;
 		} else if (!strcmp(s, "1.0")) {
@@ -1153,6 +1371,14 @@ SetScale(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		} else {
 			appData.scale = strdup(s);
 		}
+		if (appData.scale != NULL) {
+			get_scale_values(&fx, &fy);
+			if (fx <= 0.0 || fy <= 0.0) {
+				appData.scale = NULL;
+				return;
+			}
+		}
+		
 		if (appData.fullScreen) {
 			fs = 1;
 			FullScreenOff();
@@ -1162,6 +1388,7 @@ SetScale(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 			FullScreenOn();
 		}
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1173,6 +1400,7 @@ SetEscapeKeys(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	if (s[0] != '\0') {
 		appData.escapeKeys = strdup(s);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void set_ycrop(int n) {
@@ -1206,6 +1434,7 @@ SetYCrop(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		n = atoi(s);
 		set_ycrop(n);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void set_scbar(int n) {
@@ -1240,6 +1469,7 @@ SetScbar(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		n = atoi(s);
 		set_scbar(n);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1249,6 +1479,121 @@ SetScaleN(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		int n = atoi(params[0]);
 		set_server_scale(n);
 	}
+	if (w || ev || params || num_params) {}
+}
+
+void UpdateQual(void) {
+	SetFormatAndEncodings();
+	UpdateSubsampButtons();
+	UpdateQualSlider();
+}
+
+extern double latency;
+
+static void LosslessRefresh(void) {
+	String encodings = appData.encodingsString;
+	int compressLevel = appData.compressLevel;
+	int qual = appData.qualityLevel;
+	Bool enableJPEG = appData.enableJPEG;
+	appData.qualityLevel = -1;
+	appData.enableJPEG = False;
+	appData.encodingsString = "tight copyrect";
+	appData.compressLevel = 1;
+	SetFormatAndEncodings();
+	SendFramebufferUpdateRequest(0, 0, si.framebufferWidth, si.framebufferHeight, False);
+	if (latency > 0.0) {
+		if (0) usleep((int) latency * 1000);
+	}
+	appData.qualityLevel = qual;
+	appData.enableJPEG = enableJPEG;
+	appData.encodingsString = encodings;
+	appData.compressLevel = compressLevel;
+	SetFormatAndEncodings();
+}
+
+static void QualHigh(void) {
+	appData.encodingsString = "tight copyrect";
+	if(appData.useBGR233 || appDataNew.useBGR565) {
+		fprintf(stderr, "WARNING: Cannot enable JPEG because BGR233/BGR565 is enabled.\n");
+	} else {
+		appData.enableJPEG = True;
+	}
+	appData.subsampLevel = TVNC_1X;
+	appData.qualityLevel = 95;
+	UpdateQual();
+}
+
+static void QualMed(void) {
+	appData.encodingsString = "tight copyrect";
+	if(appData.useBGR233 || appDataNew.useBGR565) {
+		fprintf(stderr, "WARNING: Cannot enable JPEG because BGR233/BGR565 is enabled.\n");
+	} else {
+		appData.enableJPEG = True;
+	}
+	appData.subsampLevel = TVNC_2X;
+	appData.qualityLevel = 80;
+	UpdateQual();
+}
+
+static void QualLow(void) {
+	appData.encodingsString = "tight copyrect";
+	if(appData.useBGR233 || appDataNew.useBGR565) {
+		fprintf(stderr, "WARNING: Cannot enable JPEG because BGR233/BGR565 is enabled.\n");
+	} else {
+		appData.enableJPEG = True;
+	}
+	appData.subsampLevel = TVNC_4X;
+	appData.qualityLevel = 30;
+	UpdateQual();
+}
+
+static void QualLossless(void) {
+	appData.encodingsString = "tight copyrect";
+	appData.enableJPEG = False;
+	appData.compressLevel = 0;
+	UpdateQual();
+}
+
+static void QualLosslessWAN(void) {
+	appData.encodingsString = "tight copyrect";
+	appData.enableJPEG = False;
+	appData.compressLevel = 1;
+	UpdateQual();
+}
+
+void
+SetTurboVNC(Widget w, XEvent *ev, String *params, Cardinal *num_params)
+{
+	if (*num_params != 0) {
+		int n = atoi(params[0]);
+		if (0) fprintf(stderr, "SetTurboVNC: %d\n", n);
+		if (n == 1) {
+			QualHigh();
+		} else if (n == 2) {
+			QualMed();
+		} else if (n == 3) {
+			QualLow();
+		} else if (n == 4) {
+			QualLossless();
+		} else if (n == 5) {
+			QualLosslessWAN();
+		} else if (n == 6) {
+			appData.subsampLevel = TVNC_1X;
+			UpdateQual();
+		} else if (n == 7) {
+			appData.subsampLevel = TVNC_2X;
+			UpdateQual();
+		} else if (n == 8) {
+			appData.subsampLevel = TVNC_4X;
+			UpdateQual();
+		} else if (n == 9) {
+			appData.subsampLevel = TVNC_GRAY;
+			UpdateQual();
+		} else if (n == 10) {
+			LosslessRefresh();
+		}
+	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1258,6 +1603,7 @@ SetQuality(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		int n = atoi(params[0]);
 		set_server_quality(n);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1267,6 +1613,7 @@ SetCompress(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 		int n = atoi(params[0]);
 		set_server_compress(n);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1316,6 +1663,7 @@ SetViewOnlyState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1326,6 +1674,7 @@ SetNOJPEGState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, True, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1339,6 +1688,7 @@ SetQualityState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 			XtVaSetValues(w, XtNstate, False, NULL);
 		}
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1352,6 +1702,7 @@ SetCompressState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 			XtVaSetValues(w, XtNstate, False, NULL);
 		}
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1365,6 +1716,7 @@ SetScaleNState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 			XtVaSetValues(w, XtNstate, False, NULL);
 		}
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1381,6 +1733,7 @@ Set8bppState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1397,6 +1750,7 @@ Set16bppState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1413,6 +1767,7 @@ SetFullColorState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 			XtVaSetValues(b16, XtNstate, False, NULL);
 		}
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1423,6 +1778,7 @@ SetXGrabState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1433,6 +1789,7 @@ SetEscapeKeysState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1449,6 +1806,7 @@ Set256ColorsState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1465,6 +1823,7 @@ Set64ColorsState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1481,6 +1840,7 @@ Set8ColorsState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1491,92 +1851,79 @@ SetGreyScaleState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
+}
+
+static void init_state(void) {
+	static int first = 1;
+	if (first && appData.encodingsString) {
+		char *t, *z, *y, *h;
+		char *str = appData.encodingsString;
+		int len = strlen(str);
+
+		t = strstr(str, "tight");
+		z = strstr(str, "zrle");
+		y = strstr(str, "zywrle");
+		h = strstr(str, "hextile");
+
+		if (!t) t = str + len;
+		if (!z) z = str + len;
+		if (!y) y = str + len;
+		if (!h) h = str + len;
+
+		usingZRLE = False;
+		usingZYWRLE = False;
+		usingHextile = False;
+
+		if (t < z && t < y && t < h) {
+			;
+		} else if (z < t && z < y && z < h) {
+			usingZRLE = True;
+		} else if (y < t && y < z && y < h) {
+			usingZYWRLE = True;
+			usingZRLE = True;
+		} else if (h < t && h < z && h < y) {
+			usingHextile = True;
+		}
+	}
+	first = 0;
+
 }
 
 void
 SetZRLEState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 {
-	static int first = 1;
-	if (first && appData.encodingsString) {
-		char *t, *z, *w;
-		t = strstr(appData.encodingsString, "tight");
-		z = strstr(appData.encodingsString, "zrle");
-		w = strstr(appData.encodingsString, "zywrle");
-		if (t) {
-			if (z) {
-				if (w) {
-					if (t < z && t < w) {
-						usingZRLE = False;
-					} else {
-						usingZRLE = True;
-					}
-					if (z < w) {
-						usingZYWRLE = False;
-					} else {
-						usingZYWRLE = True;
-					}
-				} else {
-					if (t < z) {
-						usingZRLE = False;
-					} else {
-						usingZRLE = True;
-					}
-					usingZYWRLE = False;
-				}
-			} else {
-				if (w) {
-					if (t < w) {
-						usingZRLE = False;
-					} else {
-						usingZRLE = True;
-					}
-					usingZYWRLE = True;
-				} else {
-					usingZRLE = False;
-					usingZYWRLE = False;
-				}
-			}
-		} else {
-			if (z) {
-				if (w) {
-					usingZRLE = True;
-					if (z < w) {
-						usingZYWRLE = False;
-					} else {
-						usingZYWRLE = True;
-					}
-				} else {
-					usingZRLE = True;
-					usingZYWRLE = False;
-				}
-			} else {
-				if (w) {
-					usingZRLE = True;
-					usingZYWRLE = True;
-				} else {
-					usingZRLE = False;
-					usingZYWRLE = False;
-				}
-			}
-		}
-	}
-	first = 0;
-
+	init_state();
 	if (usingZRLE) {
 		XtVaSetValues(w, XtNstate, True, NULL);
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
+}
+
+void
+SetHextileState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
+{
+	init_state();
+	if (usingHextile) {
+		XtVaSetValues(w, XtNstate, True, NULL);
+	} else {
+		XtVaSetValues(w, XtNstate, False, NULL);
+	}
+	if (w || ev || params || num_params) {}
 }
 
 void
 SetZYWRLEState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 {
+	init_state();
 	if (usingZYWRLE) {
 		XtVaSetValues(w, XtNstate, True, NULL);
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1587,6 +1934,7 @@ SetCursorShapeState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1597,6 +1945,7 @@ SetCursorAlphaState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1607,6 +1956,7 @@ SetX11CursorState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1617,6 +1967,7 @@ SetBellState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, True, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1627,6 +1978,7 @@ SetRawLocalState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1637,6 +1989,40 @@ SetServerInputState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
+}
+
+void
+SetPipelineUpdates(Widget w, XEvent *ev, String *params, Cardinal *num_params)
+{
+	if (!appData.pipelineUpdates) {
+		XtVaSetValues(w, XtNstate, False, NULL);
+	} else {
+		XtVaSetValues(w, XtNstate, True, NULL);
+	}
+	if (w || ev || params || num_params) {}
+}
+
+void
+SetSendClipboard(Widget w, XEvent *ev, String *params, Cardinal *num_params)
+{
+	if (!appData.sendClipboard) {
+		XtVaSetValues(w, XtNstate, False, NULL);
+	} else {
+		XtVaSetValues(w, XtNstate, True, NULL);
+	}
+	if (w || ev || params || num_params) {}
+}
+
+void
+SetSendAlways(Widget w, XEvent *ev, String *params, Cardinal *num_params)
+{
+	if (!appData.sendAlways) {
+		XtVaSetValues(w, XtNstate, False, NULL);
+	} else {
+		XtVaSetValues(w, XtNstate, True, NULL);
+	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1647,6 +2033,7 @@ SetSingleWindowState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1657,6 +2044,7 @@ SetTextChatState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
 
 void
@@ -1667,4 +2055,5 @@ SetFileXferState(Widget w, XEvent *ev, String *params, Cardinal *num_params)
 	} else {
 		XtVaSetValues(w, XtNstate, False, NULL);
 	}
+	if (w || ev || params || num_params) {}
 }
